%!octave 
%
%    Restricted Boltzmann machine based associative memory demo
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: rbm_training.m,v 1.13.1.12 2016/03/12 17:53:58 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% алгоритм contrastive divergence обучения RBM
%
% X - учебный набор
% W - матрица весов
% er - история изменения ошибки обучения
%
function [ W,er ] = rbm_training(L,C,W,er_min,max_iter),
   ex_numb = size(L,2); % количество образов
   %batch_size = min([7,ex_numb]) ; % mini-batch: количество примеров в случайной выборке
   batch_size = ex_numb ; % full-batch: полный набор 

   er = loss(C,W) ;
   printf("[i] %i/%i : %f / %f      \r", 0, max_iter, er, er_min );
   if stop_condition(er,er_min), return; endif % условия остановки обучения


   dW = {0.0,0.0,0.0}; % начальное значение изменения весов 

   for iter=1:max_iter,
   	  % случайный пакет из учебных примеров
      batch_idx = randperm(ex_numb)(1:batch_size); 
	  Lr = L(:,batch_idx) ;

	  g = grad(Lr,W) ; % градиент ф-ции энергии
	  dW = weight_change(g,dW,W); % изменение весов
	  W = weight_correct(W,dW) ;

	  er(end+1) = loss(C,W) ; % вычисляем ошибку

	  if mod(iter,10)<1, % печатаем состояние каждые N эпох 
		 printf("[i] %i/%i : %f / %f    \r", iter, max_iter, er(end), er_min );
      endif

	  if stop_condition(er,er_min), break; endif % условия остановки обучения

   endfor
   printf("[i] %i/%i : %f / %f    \n", iter, max_iter, er(end), er_min );

   if iter >= max_iter, printf("[w] достигнут порог максимального количества циклов обучения                \n"); endif

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% градиент функции энергии
%
function retval = grad(L,W),
	  
   [V,H] = sampling(L,W); % вычислить мат.ожидание слоёв (sampling)

   gwn = ( V(:,:,1)*H(:,:,1)' - V(:,:,2)*H(:,:,2)' )' ;
   gbv = sum( H(:,:,1) - H(:,:,2) , 2 ) ;
   gbh = sum( V(:,:,1) - V(:,:,2) , 2 ) ;

   retval={ gwn, gbv, gbh } ;

   % ng = norm( [ gwn(:) ; gbv ; gbh ]);
   % retval = { gwn/ng, gbv/ng, gbh/ng } ;

   % ex_numb = size(V(:,:,1),2);
   % retval = { gwn, gbv/ex_numb, gbh/ex_numb } ;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% изменение весов
%
function retval = weight_change(g,dW,W),
   r = 0.1 ; % параметр регуляризации
   m = 0.9 ; % параметр момента 

   dWn={};
   dWn{1} = ( g{1} + r*(W{1})  ) + m*dW{1} ;
   dWn{2} = g{2} + m*dW{2} ;
   dWn{3} = g{3} + m*dW{3} ;

   retval = dWn ;

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% скорректировать веса
%
function retval = weight_correct(W,dW),
   s = 2e-3 ; % скорость обучения 
   
   Wn={};
   for i=1:length(W), 
	  Wn{i} = W{i} + dW{i}*s; 
   endfor
   
   retval = Wn ;

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = loss(X,W),
   [V H] = rbm_step(X,W) ;
   retval = msqe(X,V) ;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function  [ V, H ] = sampling(X,W),
   n = 2 ; % количество циклов sampling

   V(:,:,1) = X ; % сохраняем начальное состояние первого слоя
   [ V0 H0 ] = rbm_step(X,W) ; % вычисляем состояние второго слоя
   H(:,:,1) = H0 ; % сохраняем состояние второго слоя

   % выполняем n циклов сети
   for i=1:n, [V0 H0] = rbm_step(V0,W) ; endfor

   V(:,:,2) = V0 ; % сохраняем конечное состояние первого слоя
   [V0 H0] = rbm_step(V0,W) ; % вычисляем состояние второго слоя
   H(:,:,2) = H0 ; % сохраняем конечное состояние второго слоя

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% условия остановки обучения
% 
% er - текущая ошибка
% max_iter - максимальное количество циклов обучения
% er_min - порог допустимой ошибки
%
function retval=stop_condition(er,er_min),
   if isnan(er(end)), 
	  printf("\n[e] ошибка вычислений           \n") ;
	  exit -1;	  
	  return;
   endif

   if er(end)<er_min, 
	  printf("\n[i] достигнут порог допустимой ошибки                                       \n");
	  retval=1;
	  return;
   endif
   retval=0;
endfunction 


